/*
 * k8temp.c - Part of lm_sensors, Linux kernel modules for hardware
 *               monitoring
 * Copyright (C) 2006  Rudolf Marek <r.marek@sh.cvut.cz>
 *
 * Inspired from the w83785 and amd756 driver.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/jiffies.h>
#include <linux/pci.h>
#include <linux/hwmon.h>
#include <linux/hwmon-sysfs.h>
#include <linux/err.h>
#include <linux/mutex.h>


#define TEMP_FROM_REG(val)	(((val) - 49) * 1000)

/*
 * Functions declaration
 */


static struct k8temp_data *k8temp_update_device(struct device *dev);

struct k8temp_data {
	struct class_device *class_dev;
	struct mutex update_lock;
	const char *name;
	char valid; /* zero until following fields are valid */
	unsigned long last_updated; /* in jiffies */

	/* registers values */
	u8 sensors; /* bit6 place can be changed, bit2 core */
	u8 temp[2][2]; /* core, place */
};

static struct k8temp_data *k8temp_update_device(struct device *dev);

/*
 * Sysfs stuff
 */

static ssize_t show_name(struct device *dev, struct device_attribute
			 *devattr, char *buf)
{
	struct k8temp_data *data = dev_get_drvdata(dev);

	return sprintf(buf, "%s\n", data->name);
}


static ssize_t show_temp(struct device *dev, struct device_attribute *devattr,
	char *buf)
{
	struct sensor_device_attribute_2 *attr = to_sensor_dev_attr_2(devattr);
	int core = attr->nr;
	int place = attr->index;

	struct k8temp_data *data = k8temp_update_device(dev);
	return sprintf(buf, "%d\n", TEMP_FROM_REG(data->temp[core][place]));
}

/* core, place */

static SENSOR_DEVICE_ATTR_2(temp1_input, S_IRUGO, show_temp, NULL, 0, 0);
static SENSOR_DEVICE_ATTR_2(temp2_input, S_IRUGO, show_temp, NULL, 0, 1);
static SENSOR_DEVICE_ATTR_2(temp3_input, S_IRUGO, show_temp, NULL, 1, 0);
static SENSOR_DEVICE_ATTR_2(temp4_input, S_IRUGO, show_temp, NULL, 1, 1);
static DEVICE_ATTR(name, S_IRUGO, show_name, NULL);

static struct k8temp_data *k8temp_update_device(struct device *dev)
{
	struct k8temp_data *data = dev_get_drvdata(dev);
	struct pci_dev *pdev = to_pci_dev(dev);
	mutex_lock(&data->update_lock);

	if (!data->valid || time_after(jiffies, data->last_updated + HZ * 2)) {
		u8 tmp;

		dev_dbg(&pdev->dev, "Updating k8temp data.\n");

		pci_read_config_byte(pdev, 0xe4, &tmp);
		tmp &= ~0x44; /* Select sensor 0, core0 */
		pci_write_config_byte(pdev, 0xe4, tmp);
		pci_read_config_byte(pdev, 0xe6, &data->temp[0][0]);

		if (data->sensors & 0x40) {
			tmp |= 0x40;  /* Select sensor 1, core0 */
			pci_write_config_byte(pdev, 0xe4, tmp);
			pci_read_config_byte(pdev, 0xe6, &data->temp[0][1]);
		}

		if (data->sensors & 0x4) { 
			tmp &= ~0x40;  /* Select sensor 0, core1 */
			tmp |= 0x4;
			pci_write_config_byte(pdev, 0xe4, tmp);
			pci_read_config_byte(pdev, 0xe6, &data->temp[1][0]);
	
			if (data->sensors & 0x40) {
				tmp |= 0x40;  /* Select sensor 1, core1 */
				pci_write_config_byte(pdev, 0xe4, tmp);
				pci_read_config_byte(pdev, 0xe6, &data->temp[1][1]);
			}
		}

		data->last_updated = jiffies;
		data->valid = 1;
	}

	mutex_unlock(&data->update_lock);

	return data;
}

static struct pci_device_id k8temp_ids[] = {
	{PCI_VENDOR_ID_AMD, 0x1103, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{ 0, }
};

static int __devinit k8temp_probe(struct pci_dev *pdev,
				  const struct pci_device_id *id)
{
int err;
u8 tmp, temp;
struct k8temp_data *data; 
u32 cpuid = cpuid_eax(0x1);
	
	/* this feature should be available since SH-C0 */
	if ((cpuid == 0xf40) || (cpuid == 0xf50) || (cpuid == 0xf51)) {
		err = -ENODEV;
		goto exit;
	}

	if (!(data = kzalloc(sizeof(struct k8temp_data), GFP_KERNEL))) {
		err = -ENOMEM;
		goto exit;
	}

	pci_read_config_byte(pdev, 0xe4, &tmp);
	tmp &= ~0x44; /* Select sensor 0, core0 */
	pci_write_config_byte(pdev, 0xe4, tmp);
	pci_read_config_byte(pdev, 0xe4, &tmp);

	if (tmp & 0x44) {
		dev_err(&pdev->dev, "Configuration bits stuck at 1!\n");
		err = -ENODEV;
		goto exit_free;
	}

	tmp |= 0x44;
	pci_write_config_byte(pdev, 0xe4, tmp);

	/* now we know if we can change core and/or sensor */
	pci_read_config_byte(pdev, 0xe4, &data->sensors);

	if (data->sensors & 0x40) {
		tmp &= ~0x4; /* Select sensor 1, core0 */
		pci_write_config_byte(pdev, 0xe4, tmp);
		pci_read_config_byte(pdev, 0xe6, &temp);
		if (!temp) /* if temp is 0 -49C is not likely */
			data->sensors &= ~0x40;
	}

	if (data->sensors & 0x4) {
		tmp &= ~0x40; /* Select sensor 0, core1 */
		tmp |= 0x4;
		pci_write_config_byte(pdev, 0xe4, tmp);
		pci_read_config_byte(pdev, 0xe6, &temp);
		if (!temp) /* if temp is 0 -49C is not likely */
			data->sensors &= ~0x4;
	}

	data->name = "k8temp";
	mutex_init(&data->update_lock);
	dev_set_drvdata(&pdev->dev, data);

	/* Register sysfs hooks */
	data->class_dev = hwmon_device_register(&pdev->dev);

	if (IS_ERR(data->class_dev)) {
		err = PTR_ERR(data->class_dev);
		goto exit_free;
	}
	
	device_create_file(&pdev->dev,
			   &sensor_dev_attr_temp1_input.dev_attr);

	/* sensor can be changed and reports something*/
	if (data->sensors & 0x40)
		device_create_file(&pdev->dev,
				   &sensor_dev_attr_temp2_input.dev_attr);

	/* core can be changed and reports something */
	if (data->sensors & 0x4) { 
		device_create_file(&pdev->dev,
				   &sensor_dev_attr_temp3_input.dev_attr);

		if (data->sensors & 0x40) 
			device_create_file(&pdev->dev,
				   &sensor_dev_attr_temp4_input.dev_attr);	
	}

	device_create_file(&pdev->dev,
			   &dev_attr_name);

	return 0;

exit_free:
	dev_set_drvdata(&pdev->dev, NULL);
	kfree(data);
exit:
	return err;

}

static void __devexit k8temp_remove(struct pci_dev *pdev)
{
	struct k8temp_data *data = dev_get_drvdata(&pdev->dev);

	hwmon_device_unregister(data->class_dev);
	device_remove_file(&pdev->dev,
			   &sensor_dev_attr_temp1_input.dev_attr);
	device_remove_file(&pdev->dev,
			   &sensor_dev_attr_temp2_input.dev_attr);
	device_remove_file(&pdev->dev,
			   &sensor_dev_attr_temp3_input.dev_attr);
	device_remove_file(&pdev->dev,
			   &sensor_dev_attr_temp4_input.dev_attr);
	device_remove_file(&pdev->dev,
			   &dev_attr_name);

	dev_set_drvdata(&pdev->dev, NULL);
	kfree(data);
}

static struct pci_driver k8temp_driver = {
	.name		= "k8temp",
	.id_table	= k8temp_ids,
	.probe		= k8temp_probe,
	.remove		= __devexit_p(k8temp_remove),
};

static int __init k8temp_init(void)
{
	return pci_module_init(&k8temp_driver);
}

static void __exit k8temp_exit(void)
{
	pci_unregister_driver(&k8temp_driver);
}

MODULE_AUTHOR("Rudolf Marek <r.marek@sh.cvut.cz>");
MODULE_DESCRIPTION("k8temp driver");
MODULE_LICENSE("GPL");

module_init(k8temp_init)
module_exit(k8temp_exit)
